/*
 * Code to drive a neopixel strip from an Apple II
 * By Joe Strosnider
 * Designed for Arduino Uno - your mileage may vary.
*/


#include <Adafruit_NeoPixel.h>
#include <EEPROM.h>

//define NeoPixel Pin and Number of LEDs
    #define PIN 9
    #define NUM_LEDS 60

//create a NeoPixel strip
    Adafruit_NeoPixel strip = Adafruit_NeoPixel(NUM_LEDS, PIN, NEO_GRB + NEO_KHZ800);

// Stuff to make Digital Read/Write much faster.
#define portOfPin(P)\
  (((P)>=0&&(P)<8)?&PORTD:(((P)>7&&(P)<14)?&PORTB:&PORTC))
#define ddrOfPin(P)\
  (((P)>=0&&(P)<8)?&DDRD:(((P)>7&&(P)<14)?&DDRB:&DDRC))
#define pinOfPin(P)\
  (((P)>=0&&(P)<8)?&PIND:(((P)>7&&(P)<14)?&PINB:&PINC))
#define pinIndex(P)((uint8_t)(P>13?P-14:P&7))
#define pinMask(P)((uint8_t)(1<<pinIndex(P)))

#define pinAsInput(P) *(ddrOfPin(P))&=~pinMask(P)
#define pinAsInputPullUp(P) *(ddrOfPin(P))&=~pinMask(P);digitalHigh(P)
#define pinAsOutput(P) *(ddrOfPin(P))|=pinMask(P)
#define digitalLow(P) *(portOfPin(P))&=~pinMask(P)
#define digitalHigh(P) *(portOfPin(P))|=pinMask(P)
#define isHigh(P)((*(pinOfPin(P))& pinMask(P))>0)
#define isLow(P)((*(pinOfPin(P))& pinMask(P))==0)
#define digitalState(P)((uint8_t)isHigh(P))

/*******************/
/* Variables Setup */
/*******************/

// Colors
    const uint32_t  green = strip.Color(32, 255, 0);
    const uint32_t  yellow = strip.Color(255, 132, 0);
    const uint32_t  orange = strip.Color(255, 40, 0);
    const uint32_t  red = strip.Color(255, 0, 0);
    const uint32_t  violet = strip.Color(255, 0, 192);
    const uint32_t  blue = strip.Color(4, 48, 255);
    const uint32_t  white = strip.Color(255, 255, 255);
    const uint32_t  black = strip.Color(0, 0, 0);
    uint32_t ColorArray[NUM_LEDS+1]; //= {green,green,green,green,green,green,green,green,green,green,yellow,yellow,yellow,yellow,yellow,yellow,yellow,yellow,yellow,yellow,orange,orange,orange,orange,orange,orange,orange,orange,orange,orange,red,red,red,red,red,red,red,red,red,red,violet,violet,violet,violet,violet,violet,violet,violet,violet,violet,blue,blue,blue,blue,blue,blue,blue,blue,blue,blue,green};
    byte EEPromArray[NUM_LEDS*4];

// Pins
    const byte interruptPin = 2;
    const byte A2ClockPin = 3;
    const byte A2DataPin = 4;
    const byte A2CommandPin = 5;

 // Modes
    volatile byte state = 0;
    volatile byte SaveState = 0;
    const byte ModeAPPLE = 0;
    const byte ModeCHASE = 1;
    const byte ModeFADE = 2;
    const byte ModeRANDOM = 3; // as long as you always put RANDOM last in the list but before ModePLAY, the interrupt code will cycle modes properly. :)
    const byte ModePLAY = 4; 
    const byte ModeCOMMAND = 255;

// Commands
    const byte CommandCLEAR = 0;
    const byte CommandFILL = 1;
    const byte CommandPIXEL = 2;
    const byte CommandCHASE = 3;
    const byte CommandFADE = 4;
    const byte CommandBUILTIN = 254;
    const byte CommandRESET = 255;
    
    const byte CommandArraySize = 61;
    byte CommandArray[61];  // maximum number of bytes in the command array.
                                          // first byte = number of commands
                                          // subsequent bytes = commands and their arguments
                                          // currently, this is sized to allow a quantity of ten commands that have five arguments each.
                                          // There is a sanity check to prevent overflow, but checking should also be done on the APPLE II ~BEFORE~ sending commands.

// eeprom locations
    const byte EEPROMModePtr = 0;
    const byte EEPROMArryPtr = 1;

// General Flags
    byte DefaultAnimation = 0;   // So we know when an animated mode has been properly set.
    unsigned long previousMillis = 0UL;
    unsigned long interval = 1000UL;

// Setup Routine: Runs once on startup/reset
    void setup() {

          Serial.begin(115200);         // Start Serial Comms for debugging
          while (!Serial) {}            // wait for serial port to connect. Needed for native USB port only

          // Ensure the Interrupt pin is on input and pulled up.
          pinMode(interruptPin, INPUT_PULLUP); 

          // Set Up The Interrupt Service Routine                                   
          attachInterrupt(digitalPinToInterrupt(interruptPin), interruptHandler, FALLING);
          
          pinMode(A2ClockPin, INPUT_PULLUP);      // Ensure the clock pin is on input and pulled up.
          pinMode(A2DataPin, INPUT_PULLUP);       // Ensure the data pin is on input and pulled up.
          pinMode(A2CommandPin, INPUT_PULLUP);    // Ensure the command pin is on input and pulled up.
          
          strip.begin();                          // start the strip 
          strip.show();                           // Blank it to zero

          EEPROM.get(EEPROMModePtr, state);       // read in the mode byte
          SaveState = state;
          
          /*
          String PrintOut = "SaveState at reset is: ";
          if (SaveState==ModeAPPLE){PrintOut=PrintOut+"ModeAPPLE";}
          if (SaveState==ModeCHASE){PrintOut=PrintOut+"ModeCHASE";}
          if (SaveState==ModeFADE){PrintOut=PrintOut+"ModeFADE";}
          if (SaveState==ModeRANDOM){PrintOut=PrintOut+"ModeRANDOM";}
          if (SaveState==ModePLAY){PrintOut=PrintOut+"ModePLAY";}
          Serial.println(PrintOut);
          */
          
          if (state == 255){                      // if the state is blank (i.e. this is a BRAND NEW arduino) then deal with that.
            state = ModeAPPLE;
            SaveState = state;
          }


          // Otherwise, read in the pixel strip data
          else {
            LoadStripFromEEPROM();
            GetStripIntoArray();
            DefaultAnimation = 1;
            strip.show();
          }
    }


// Primary Logic Loop
    void loop() {

      // Mode: APPLE
      // Proudly show the Apple Colors in a static pattern.
        while (state == ModeAPPLE){
              AppleColors();      // set the strip to Apple Colors
              state = ModePLAY;   // set the mode to play
          }


      // Mode: CHASE
      // Do a sequential chase of whatever is already in the strip.
        while (state == ModeCHASE) {
            if (DefaultAnimation == 0){       // If we get here from a mode cycle....
                  DefaultAnimation = 1;       // ... then tell ourselves as such, so we dont keep resetting the strip....
                  AppleColors();              // ... to the Apple colors.
              }
            
            GetStripIntoArray();                            // Assumes strip already has the pattern to chase.
            ColorArray[NUM_LEDS]=ColorArray[0];             // preserve first entry
            
            for (int i = 0; i <= NUM_LEDS-1; i++) {   
                ColorArray[i] = ColorArray[i+1];            // shift all the array entries
            }
            
            for (int i = 0; i <= NUM_LEDS-1; i++) {
                strip.setPixelColor(i, ColorArray[i]);      // send the array back to the strip
            }
            
            strip.show();
            delay(50);
        }

      // Mode: FADE
      // Do a fade from whatever is in the strip, down to black, and back.
        while (state == ModeFADE){
            if (DefaultAnimation == 0){     // If we get here from a mode cycle....
                  DefaultAnimation = 1;     // ... then tell ourselves as such, so we dont keep resetting the strip....
                  AppleColors();            // ... to the Apple colors.
                  SaveEEPROM();             // then save it so we don't lose it on restart
              }
            GetStripIntoArray();                                                      // Assumes the strip is already assigned and showing the pattern to fade.
            uint32_t LocalArray[NUM_LEDS+1];                                          // preserve the original array so we know what to fade back to.
            for (int i = 0; i <= NUM_LEDS; i++) {      
              LocalArray[i]=ColorArray[i];
            }

            byte steps = 255;
            
            for (int i = 0; i <= steps; i++) {                // fade down
              if (state!=ModeFADE ){break;}                   // exits early if there's a mode change.
              for (int j = 0;j<NUM_LEDS;j++){
                uint8_t R = (ColorArray[j] >> 16) & 0xFF;
                uint8_t G = (ColorArray[j] >>8) & 0xFF;
                uint8_t B = ColorArray[j] & 0xFF;
                int Z = steps - i;
                R = (R*Z) / steps;
                G = (G*Z) / steps;
                B = (B*Z) / steps;
                ColorArray[j] = strip.Color(R,G,B);
                strip.setPixelColor(j, ColorArray[j]);
              }
              strip.show();
              if (i<=48){ delay (50); }                       // deals with a weird math problem I can't figure out...
              else {break;}
            }
            
            if (state!=ModeFADE ){break;}                     // exits early if there's a mode change.
            
            for (int i = 0; i <= steps; i++) {                // fade up
              if (state!=ModeFADE ){break;}                   // exits early if there's a mode change.
              for (int j = 0;j<NUM_LEDS;j++){
                uint8_t R = (ColorArray[j] >> 16) & 0xFF;
                uint8_t G = (ColorArray[j] >>8) & 0xFF;
                uint8_t B = ColorArray[j] & 0xFF;
      
                uint8_t lR = (LocalArray[j] >> 16) & 0xFF;
                uint8_t lG = (LocalArray[j] >>8) & 0xFF;
                uint8_t lB = LocalArray[j] & 0xFF;
                
                int Z = steps - i;
                R = ((R * Z) + (lR * i)) / steps;
                G = ((G * Z) + (lG * i)) / steps;
                B = ((B * Z) + (lB * i)) / steps;
                ColorArray[j] = strip.Color(R,G,B);
                strip.setPixelColor(j, ColorArray[j]);
              }
              strip.show();
              if (i<=48){ delay (50); }                       // deals with a weird math problem I can't figure out...
              else {break;}
            }
            LoadStripFromEEPROM();      // this fixes rounding errors - reloads strip with known good value.
      }


      // Mode: RANDOM
      // Twinkle the Apple Colors on the pixels randomly
      while (state == ModeRANDOM){
          uint32_t RColor;
          for (int i = 0; i <= NUM_LEDS; i++) {
                byte r=random(0, 5);
                switch (r){
                  case 0: RColor = red; break;
                  case 1: RColor = orange; break;
                  case 2: RColor = yellow ;break;
                  case 3: RColor = green ;break;
                  case 4: RColor = blue ;break;
                  case 5: RColor = violet; break;
                  default: RColor = green; break;
                } 
             strip.setPixelColor(i, RColor);  //pixel number, Red, Green, Blue
          }
          delay (62);
          strip.show();
      }


      // Mode: PLAY
      // We have a custom static pattern. YAY!
      while (state == ModePLAY){
          // the strip has a custom static pattern. Do nothing.
        }


      // Mode: COMMAND
      // Fetches a command from the Apple II
      while (state == ModeCOMMAND){
              //Serial.println("Got ModeCOMMAND");     
              int ArryPtr=0;
              int bitPtr=0;
              while (digitalRead(A2CommandPin)==LOW){                                               // Command Pin should be low when reaching here.                      
                     while (digitalRead(A2ClockPin)==LOW){                                          // If clock isn't low yet, we loop around and wait for it to be so.
                            if (ArryPtr>CommandArraySize){break;}                                   // If the pointer is greater than the size of the array, stop reading more bits.
                            bitWrite(CommandArray[ArryPtr], bitPtr, digitalRead(A2DataPin));        // once clock goes low, assign the bit to the correct array position
                            bitPtr++;                                                               // increment the bit pointer.
                            if (bitPtr>7){
                                bitPtr=0;                                                           // if it is 8, then reset bit pointer to zero
                                ArryPtr++;                                                          // 8 also mean's we've rolled over to the next byte. Increment the array pointer.
                              }            
                            while (digitalRead(A2ClockPin)==LOW){
                              /* spinlock to wait for clock to go high again */
                              }
                     }                                                                              // if we've reached here, we've read a bit and the clock is high again.
              }                                                                                     // go check to see if the command pin is still low.
              // Serial.println("Command Fetch Finished");
              // once we reach here, the command is done because the command pin is high.
            Inject();                                                                               // Send the command to the strip


            // if we're still in COMMAND mode after having injected the strip, that means we need to go "play" the command
            if (state == ModeCOMMAND){state = ModePLAY;SaveState = state;}                          
        }
          
    }  // End Main Loop


void Inject(){          // Injects an existing Data array into the strip.
    int ArryPtr=1;      // Set the pointer to one, because the Zero'th item in the array always equals the number of commands in the array.
    
    if (!CommandArray[0]){     // handles the case that we get to Inject but there's no command in the buffer.
        Serial.print("CommandArray buffer error. Contents of CommandArray[0]: ");
        Serial.println(CommandArray[0]);
        state=SaveState;
      }

    for (int i = 1; i <= CommandArray[0]; i++){
      switch (CommandArray[ArryPtr]){
          case CommandCLEAR:      // Clear: has zero arguments.
              strip.clear();
              strip.show();
              state = ModePLAY;
              SaveState = state;
              SaveEEPROM();
              ArryPtr++;           // Pointer is always incremented by the number of arguments, plus one.
              //Serial.println("Got CommandCLEAR");
              break;
              
          case CommandFILL:      // Fill: has five Arguments
              strip.fill(strip.Color(CommandArray[ArryPtr+1], CommandArray[ArryPtr+2], CommandArray[ArryPtr+3]), CommandArray[ArryPtr+4], CommandArray[ArryPtr+5]);
              strip.show();
              state = ModePLAY;
              SaveState = state;
              SaveEEPROM();
              ArryPtr=ArryPtr+6;   // Pointer is always incremented by the number of arguments, plus one.
              //Serial.println("Got CommandFILL");
              break;
              
          case CommandPIXEL:    // Set Pixel: has four arguments
              strip.setPixelColor(CommandArray[ArryPtr+1], CommandArray[ArryPtr+2], CommandArray[ArryPtr+3], CommandArray[ArryPtr+4]);
              strip.show();
              state = ModePLAY;
              SaveState = state;
              SaveEEPROM();
              ArryPtr=ArryPtr+5;   // Pointer is always incremented by the number of arguments, plus one.
              //Serial.println("Got CommandPIXEL");
              break;
    
         case CommandCHASE:    // Chase: has zero arguments - Should always be the last command. Assumes you've already loaded the strip with the pattern to chase.
              ArryPtr++;
              DefaultAnimation = 1;
              state = ModeCHASE;
              SaveState = state;
              SaveEEPROM();
              //Serial.println("Got CommandCHASE");
              break;
              
         case CommandFADE:    // Fade: has zero arguments - Should always be the last command. Assumes you've already loaded the strip with the pattern to fade.
              ArryPtr++;
              DefaultAnimation = 1;
              state = ModeFADE;
              SaveState = state;
              SaveEEPROM();
              //Serial.println("Got CommandFADE");
              break;
              
         case CommandBUILTIN:    // Set to one of the built in Apple Color modes.
              //Serial.println("Got CommandBUILTIN");
              DefaultAnimation = 0;
              state = CommandArray[ArryPtr+1];
              SaveState = state;
              ArryPtr++;
              SaveEEPROM();
              break;
              
         case CommandRESET:    // Reset: has zero arguments
              //Serial.println("Got CommandRESET");
              DefaultAnimation = 0;
              ArryPtr++;
              state = ModeAPPLE;
              SaveState = state;
              SaveEEPROM();
              break;       
                              
         default:
              Serial.print("COMMAND ERROR: ");
              Serial.println(CommandArray[ArryPtr]);
              break;     // We fell through due to a command error.
      }
    }
  }

void LoadStripFromEEPROM(){
  for (int i = 0; i <= ((NUM_LEDS*4)-1); i++) {
      EEPROM.get(EEPROMArryPtr+i, EEPromArray[i]); //b,g,r,a
      if ((i+1)%4==0){
        int CurrentPixel = (i-3)/4;
        strip.setPixelColor(CurrentPixel, EEPromArray[i-1],EEPromArray[i-2],EEPromArray[i-3],EEPromArray[i]);
      }
  }  
}


void SaveEEPROM(){
    //Serial.println("Got to SaveEEPROM");
    GetStripIntoArray();                              // ensures we have the most up-to-date strip value
    EEPROM.put(EEPROMModePtr, SaveState);                // save the mode byte                               
    for (int i = 0; i <= (NUM_LEDS*4); i+=4) {             // save the array data
        EEPROM.put(EEPROMArryPtr+i, ColorArray[i/4]);
    }

    EEPROM.get(EEPROMModePtr, SaveState);                // save the mode byte
    /*
    String PrintOut = "State in EEPROM after Saving Is: ";
    if (SaveState==ModeAPPLE){PrintOut=PrintOut+"ModeAPPLE";}
    if (SaveState==ModeCHASE){PrintOut=PrintOut+"ModeCHASE";}
    if (SaveState==ModeFADE){PrintOut=PrintOut+"ModeFADE";}
    if (SaveState==ModeRANDOM){PrintOut=PrintOut+"ModeRANDOM";}
    if (SaveState==ModePLAY){PrintOut=PrintOut+"ModePLAY";}
    Serial.println(PrintOut);
    */
}

void AppleColors(){ // sets the strip to static Apple II Colors
    strip.fill(green, 0, 10);
    strip.fill(yellow, 10, 10);
    strip.fill(orange, 20, 10);
    strip.fill(red, 30, 10);
    strip.fill(violet, 40, 10);
    strip.fill(blue, 50, 10);
    strip.show();
 }

void GetStripIntoArray () {  // Reads the current strip into an array we can manipulate.
    for (int i=0;i<=NUM_LEDS;i++){
      ColorArray[i]=strip.getPixelColor(i);
    }
 }

void interruptHandler() {                               // Interrupt Handler  
    static unsigned long last_interrupt_time = 0;
    unsigned long interrupt_time = millis();
      if (interrupt_time - last_interrupt_time > 20){   // debounces the interrupt
            DefaultAnimation = 0;                       // Reset animation initialization. If 0, means we want to animate what is already there.
            if (!isHigh(A2CommandPin)){
              SaveState = state;                        // preserve the previous state
              state = ModeCOMMAND;                      // if the command pin is low on interrupt, then the Apple II wants to send a command.     
            }                              
      }
    last_interrupt_time = interrupt_time;
 }
